;;; @ ltx-dead.el -- dead keys for LaTeX format accented characters. 
;;; 
;;; $Id: ltx-dead.el,v 5.6 1993/07/08 17:51:45 amanda Exp $

(require 'min-bind)

;;; @@ Description
;;;
;;; This module implements dead keys for LaTeX format accented characters.
;;;
;;; Here is one way to support accents.  The following file implements
;;; accents as ``dead keys'' in emacs.  To use it, you must set the
;;; variable LaTeX-dead-keys to a list of accents you are interested in.
;;; 
;;; For example:
;;;	 (setq LaTeX-dead-keys '(?` ?\" ?^))
;;; will define `, ", and ^ as dead keys.
;;; 
;;; You must then activate the dead keys with M-x LaTeX-dead-mode.
;;; 
;;; When you press a dead key, emacs will wait for you to enter a letter.
;;; The letter will then be inserted with the appropriate accent.  If you
;;; press the dead key twice, the accent itself will be inserted.
;;; 
;;; This mode should be useful if you use relatively many accented
;;; letters.  If you only use few accents, you will probably not like it.
;;; In that case, use C-c C-m instead.

;;; @@ Customization

(defvar LaTeX-dead-keys '(?' ?` ?\" ?^ ?~)
  "*List of characters used as dead keys.")

;;; @@ Local Variables

(defvar LaTeX-dead-mode nil
  "This variable is nil unless LaTeX-dead-mode is active.")
  
  (make-variable-buffer-local 'LaTeX-dead-mode)

(or (assoc 'LaTeX-dead-mode minor-mode-alist)
    (setq minor-mode-alist (cons '(LaTeX-dead-mode " Dead") 
				 minor-mode-alist)))

;;; @@ The Mode

(defun LaTeX-dead-mode (&optional arg)
  "Toggle LaTeX-dead-mode.
With prefix arg, turn LaTeX-dead-mode on iff arg is positive.

This mode allows users to conveniently enter characters with accents,
as found in some European languages.

The supported accents are defined by LaTeX-dead-keys.  Modify that
variable to change the list of dead keys. (You most likely want to
do that).

When a dead key is pressed, the system waits for the user to enter
another character.  This character will be accented.  To enter the
accent itself, press the dead key twice.  For example:

Key  |	Text
-----+--------
' e  |  \'{e}
' '  |  '
' i  |  \'{\i}

As shown by the example, this mode knows that i and j must be treated
specially."

  (interactive "P")
  (if (or (and (null arg) LaTeX-dead-mode)
	  (<= (prefix-numeric-value arg) 0))
      (if LaTeX-dead-mode
	  (minor-unbind 'LaTeX-dead-mode))
    (if (not LaTeX-dead-mode)
	(let ((map (make-sparse-keymap)))
	  (mapcar (function (lambda (key)
			      (define-key map (char-to-string key)
				'LaTeX-dead-function)))
		  LaTeX-dead-keys)
	  (minor-add-to-keymap 'LaTeX-dead-mode map)
	  (minor-set-variable 'LaTeX-dead-mode 'LaTeX-dead-mode t))))
  (set-buffer-modified-p (buffer-modified-p)))

;;; @@ Functions

(defun LaTeX-dead-function (char)
  "Insert CHAR, and insert it with an accent if it is a letter.  If it
is the accent itself, insert it.  Otherwise, insert the accent
character followed by the other character."
  (interactive "*c")
  (cond ((eq (aref (this-command-keys) 0) char)
	 (minor-call-shadow 'LaTeX-dead-mode (char-to-string char)))
	((eq (char-syntax char) ?w)
	 (insert "\\")
	 (self-insert-command 1)
	 (insert "{")
	 (if (or (eq char ?i) (eq char ?j))
	     (insert "\\"))
	 (insert char "}"))
	(t
	 (minor-call-shadow 'LaTeX-dead-mode
			    (char-to-string (aref (this-command-keys) 0)))
	 (command-execute (char-to-string char)))))

;;; @@ Emacs

(run-hooks 'TeX-after-ltx-dead-hook)

;;; Local Variables:
;;; mode: emacs-lisp
;;; mode: outline-minor
;;; outline-regexp: ";;; @+\\|(......"
;;; End:
