;;; @ ltx-sec.el - Section management for auc-tex
;;; 
;;; $Id: ltx-sec.el,v 1.12 1993/04/09 20:52:27 amanda Exp $

(require 'tex-site)

;;; @@ Copyright
;;;
;;; Copyright (C) 1992 Per Abrahamsen
;;; 
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2, or (at your option)
;;; any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;; @@ Customization

(defvar LaTeX-section-hooks
  '(LaTeX-section-level-hook
    LaTeX-section-title-hook
;;; LaTeX-section-toc-hook		; Most people won't want this
    LaTeX-section-section-hook
    LaTeX-section-label-hook)
  "*List of hooks to run when a new section is inserted.

The following variables are set before the hooks are run

level - numeric section level, see the documentation of `LaTeX-section'.
name - name of the sectioning command, derived from `level'.
title - The title of the section, default to an empty string.
toc - Entry for the table of contents list, default nil.
done-mark - Position of point afterwards, default nil (meaning end).

The following standard hook exist -

LaTeX-section-level-hook: Query the user about the name of the
sectioning command.  Modifies `level' and `name'.

LaTeX-section-title-hook: Query the user about the title of the
section. Modifies `title'.

LaTeX-section-toc-hook: Query the user for the toc entry.  Modifies
`toc'. 

LaTeX-section-section-hook: Insert LaTeX section command according to
`name', `title', and `toc'.  If `toc' is nil, no toc entry is
enserted.  If `toc' or `title' are empty strings, `done-mark' will be
placed at the point they should be inserted.

LaTeX-section-label-hook: Insert a label after the section command.
Controled by the variable `LaTeX-section-label'.

To get a full featured LaTeX-section command, insert 

 (setq LaTeX-section-hooks
       '(LaTeX-section-level-hook
	 LaTeX-section-title-hook
	 LaTeX-section-toc-hook
	 LaTeX-section-section-hook
	 LaTeX-section-label-hook))

in your .emacs file.")

(defvar LaTeX-section-label
  '(("chapter" . "cha:")
    ("section" . "sec:")
    ("subsection" . "sec:"))
  "*Default prefix when asking for a label.

If it is a string, it it used unchanged for all kinds of sections. 
If it is nil, no label is inserted.
If it is a list, the list is searched for a member whose car is equal
to the name of the sectioning command being inserted.  The cdr is then
used as the prefix.  If the name is not found, or if the cdr is nil,
no label is inserted.")

;;; @@ Commands

(defun LaTeX-section (arg)
  "Insert a template for a LaTeX section.
Determinate the type of section to be inserted, by the argument ARG.

If ARG is nil or missing, use the current level.
If ARG is a list (selected by C-u), go downward one level.
If ARG is negative, go up that many levels.
If ARG is positive or zero, use absolute level:

  0 : part
  1 : chapter
  2 : section
  3 : subsection
  4 : subsubsection
  5 : paragraph
  6 : subparagraph

The following variables can be set to customize:

LaTeX-section-hooks	Hooks to run when inserting a section.
LaTeX-section-label	Prefix to all section labels."
  
  (interactive "*P")
  (let* ((val (prefix-numeric-value arg))
	 (level (cond ((null arg)
		       (LaTeX-current-section))
		      ((listp arg)
		       (LaTeX-down-section))
		      ((< val 0)
		       (LaTeX-up-section (- val)))
		      (t val)))
	 (name (LaTeX-section-name level))
	 (toc nil)
	 (title "")
	 (done-mark (make-marker)))
    (newline)
    (run-hooks 'LaTeX-section-hooks)
    (newline)
    (if (marker-position done-mark)
	(goto-char (marker-position done-mark)))
    (set-marker done-mark nil)))

;;; @@ Standard Hooks.

(defun LaTeX-section-level-hook ()
  "Hook to prompt for LaTeX section name.
Insert this hook into LaTeX-section-hooks to allow the user to change
the name of the sectioning command inserted with M-x LaTeX-section."
  (let ((string (completing-read
		 (concat "Select level: (default " name ") ")
		 LaTeX-section-list
		 nil nil nil)))
    ; Update name
    (if (not (zerop (length string)))
	(setq name string))
    ; Update level
    (setq level (LaTeX-section-level name))))

(defun LaTeX-section-title-hook ()
  "Hook to prompt for LaTeX section title.
Insert this hook into LaTeX-section-hooks to allow the user to change
the title of the section inserted with M-x LaTeX-section."
  (setq title (read-input "What title: ")))

(defun LaTeX-section-toc-hook ()
  "Hook to prompt for the LaTeX section entry in the table of content .
Insert this hook into LaTeX-section-hooks to allow the user to insert
a different entry for the section in the table of content."
  (setq toc (read-input "Toc Entry: "))
  (if (zerop (length toc))
      (setq toc nil)))

(defun LaTeX-section-section-hook ()
  "Hook to insert LaTeX section command into the file.
Insert this hook into LaTeX-section-hooks after those hooks which sets
the `name', `title', and `toc' variables, but before those hooks which
assumes the section already is inserted."
    (insert TeX-esc name)
    (cond ((null toc))
	  ((zerop (length toc))
	   (insert LaTeX-optop)
	   (set-marker done-mark (point))
	   (insert LaTeX-optcl))
	  (t
	   (insert LaTeX-optop toc LaTeX-optcl)))
    (insert TeX-grop)
    (if (zerop (length title))
	(set-marker done-mark (point)))
    (insert title TeX-grcl)
    (newline))

(defun LaTeX-section-label-hook ()
  "Hook to insert a label after the sectioning command.
Insert this hook into LaTeX-section-hooks to prompt for a label to be
inserted after the sectioning command.

The beaviour of this hook is controled by LaTeX-section-label."
  (let ((prefix (cond ((null LaTeX-section-label)
		       nil)
		      ((stringp LaTeX-section-label)
		       LaTeX-section-label)
		      ((assoc level LaTeX-section-label)
		       (cdr (assoc level LaTeX-section-label)))
		      ((assoc name LaTeX-section-label)
		       (cdr (assoc name LaTeX-section-label))))))
    (if prefix
	(let ((label (read-input "What label: " prefix)))
	  (if (string-equal prefix label)
	      ()				; Done - no label entered
	    (insert TeX-esc "label" TeX-grop label TeX-grcl)
	    (LaTeX-add-labels label)
	    (newline))))))

;;; @@ Private Functions

(defun LaTeX-current-section ()
  "Return the level of the section that contain point.
See also LaTeX-section for description of levels."
  (save-excursion
    (max (LaTeX-largest-level)
	 (if (re-search-backward (LaTeX-outline-regexp) nil t)
	     (- (LaTeX-outline-level) (LaTeX-outline-offset))
	   (LaTeX-largest-level)))))

(defun LaTeX-down-section ()
  "Return the value of a section one level under the current. Tries to
find what kind of section that have been used earlier in the text, If
this fail, it will just return one less than the current section."
  (save-excursion 
    (let ((current (LaTeX-current-section))
	  (next nil)
	  (regexp (LaTeX-outline-regexp)))
      (if (not (re-search-backward regexp nil t))
	  (1+ current)
	(while (not next)
	  (cond
	   ((eq (LaTeX-current-section) current)
	    (if (re-search-forward regexp nil t)
		(if (<= (setq next (LaTeX-current-section)) current) ;Wow!
		    (setq next (1+ current)))
	      (setq next (1+ current))))
	   ((not (re-search-backward regexp nil t))
	    (setq next (1+ current)))))
	next))))

(defun LaTeX-up-section (arg)
  "Return the value of the section ARG levels above this one."
  (save-excursion
    (if (zerop arg)
	(LaTeX-current-section)
      (let ((current (LaTeX-current-section)))
	(while (and (>= (LaTeX-current-section) current)
		    (re-search-backward (LaTeX-outline-regexp)
					nil t)))
	(LaTeX-up-section (1- arg))))))


;;; @@ Emacs

(run-hooks 'TeX-after-ltx-sec-hook)

;;; Local Variables:
;;; mode: emacs-lisp
;;; mode: outline-minor
;;; outline-regexp: ";;; @+\\|(......"
;;; End:
