;;; @ tex-init.el - Support for TeX and LaTeX documents.
;;; 
;;; $Id: tex-init.el,v 5.23 1993/07/28 22:31:39 amanda Exp $

(provide 'tex-init)
(require 'auc-ver)
(require 'tex-site)

;;; @@ Copyright
;;;
;;; Copyright (C) 1985, 1986 Free Software Foundation, Inc.
;;; Copyright (C) 1987 Lars Peter Fischer
;;; Copyright (C) 1991 Kresten Krab Thorup
;;; Copyright (C) 1993 Per Abrahamsen 
;;; 
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 1, or (at your option)
;;; any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;; @@ Import
;;;
;;; The following are autoloads for different modules, as well as
;;; variables that has to be made buffer local and functions we really
;;; need now.

(defvar no-doc
  "This function is part of AUC TeX, but has not yet been loaded.
Full documentation will be available after autoloading the function."
  "Documentation for autoload functions.")

(autoload 'LaTeX-bibliography-list "ltx-misc" no-doc nil)
(autoload 'LaTeX-section "ltx-sec" no-doc t)
(autoload 'TeX-help-error "tex-buf" no-doc nil)
(autoload 'TeX-complete-symbol "tex-cpl" no-doc t)
(autoload 'TeX-insert-macro "tex-cpl" no-doc t)
(autoload 'TeX-auto-write-hook "tex-auto" no-doc nil)
(autoload 'TeX-auto-find-files "tex-auto" no-doc nil)
(autoload 'TeX-auto-store-self "tex-auto" no-doc t)

(autoload 'TeX-comment-region "tex-misc" no-doc t)
(autoload 'TeX-comment-paragraph "tex-misc" no-doc t)
(autoload 'TeX-comment-out-region "tex-misc" no-doc t)
(autoload 'TeX-comment-out-paragraph "tex-misc" no-doc t)
(autoload 'TeX-un-comment "tex-misc" no-doc t)
(autoload 'TeX-un-comment-region "tex-misc" no-doc t)

;;; @@ Buffer

(autoload 'TeX-home-buffer "tex-buf" no-doc t)
(autoload 'TeX-command-region "tex-buf" no-doc t)
(autoload 'TeX-command-master "tex-buf" no-doc t)
(autoload 'TeX-command "tex-buf" no-doc nil)
(autoload 'TeX-kill-job "tex-buf" no-doc t)
(autoload 'TeX-recenter-output-buffer "tex-buf" no-doc t)
(autoload 'TeX-next-error "tex-buf" no-doc t)
(autoload 'TeX-toggle-debug-boxes "tex-buf" no-doc t)

(defvar TeX-trailer-start nil
  "Regular expression delimiting start of trailer in a TeX file.")

 (make-variable-buffer-local 'TeX-trailer-start)

(defvar TeX-header-end nil
  "Regular expression delimiting end of header in a TeX file.")

 (make-variable-buffer-local 'TeX-header-end)

(defvar TeX-command-default nil
  "The default command for TeX-command in the current major mode.")

 (make-variable-buffer-local 'TeX-command-default)


;;; @@ Master File

(defvar TeX-one-master "\\.tex$"
  "*Regular expression matching ordinary TeX files.

You should set this variable to match the name of all files, where
automatically adding a file variable with the name of the master file
is a good idea.  When AUC TeX add the name of the master file as a
file variable, it does not need to ask next time you edit the file.  

If you dislike AUC TeX automatically modifying your files, you can set
this variable to \"<none>\".")

(defun TeX-master-file (&optional extension)
  "Return the name of the master file for the current document.

If optional argument EXTENSION is non-nil, add that file extension to
the name.

Currently is will check for the presence of a ``Master:'' line in
the beginning of the file, but that feature will be phased out."
  (let ((my-name (if (buffer-file-name)
                     (TeX-strip-extension nil (list TeX-default-extension) t)
                   "<none>")))
    (save-excursion
      (goto-char (point-min))
      (cond
       ;; Special value 't means it is own master (a free file).
       ((eq t TeX-master)
        (setq TeX-master my-name))
       
       ;; For files shared between many documents.
       ((eq 'shared TeX-master)
        (setq TeX-master
              (TeX-strip-extension
               (read-file-name "Master file: (default this file) "
                               nil "///")
               (list TeX-default-extension)
               t))
        (if (or (string-equal TeX-master "///")
                (string-equal TeX-master ""))
            (setq TeX-master my-name)))
       
       ;; We might already know the name.
       (TeX-master)
       
       ;; Support the ``Master:'' line (under protest!)
       ((re-search-forward
         "^%% *[Mm]aster:?[ \t]*\\([^ \t\n]+\\)" 500 t)
        (setq TeX-master
              (TeX-strip-extension (TeX-match-buffer 1)
                                   (list TeX-default-extension)))
        (if TeX-convert-master
            (progn
              (beginning-of-line)
              (kill-line 1)
              (TeX-add-local-master))))
       
       ;; Is this a master file?
       ((re-search-forward TeX-header-end nil t)
        (setq TeX-master my-name))
       
       ;; Ask the user (but add it as a local variable).
       (t
        (setq TeX-master
              (TeX-strip-extension
               (read-file-name "Master file: (default this file) "
                               nil "///")
               (list TeX-default-extension)
               t))
        (if (or (string-equal TeX-master "///")
                (string-equal TeX-master ""))
            (progn
              (setq TeX-master t)
              (TeX-add-local-master)
              (setq TeX-master my-name))
          (TeX-add-local-master))))))
  
  (cond ((string-match "^\\([^.].*\\)\\.[^.]*$" TeX-master)
	 ;; For files using the "tex" extension, TeX-master will be the
	 ;; filename without the extension.  We assume in the following
	 ;; that such files contain no additional dots.
	 (let ((master (substring TeX-master
				  (match-beginning 1) (match-end 1))))
	   (cond ((null extension)
		  master)
		 ((string-equal extension TeX-default-extension)
		  TeX-master)
		 (t
		  (concat master "." extension)))))
	((null extension)
	 TeX-master)
	(t
	 (concat TeX-master "." extension))))

(defvar TeX-master nil
  "*The master file associated with the current buffer.
If the file being edited is actually included from another file, you
can tell AUC TeX the name of the master file by setting this variable.
If there are multiple levels of nesting, specify the top level file. 

If this variable is nil, AUC TeX will query you for the name.

If the variable is t, AUC TeX will assume the file is a mater file
itself.

If the variable is 'shared, AUC TeX will query for the name, but not
change the file.  

It is suggested that you use the File Variables (see the info node in
the Emacs manual) to set this variable permanently for each file.")

 (make-variable-buffer-local 'TeX-master)

(defvar TeX-convert-master t
  "*If not nil, automatically convert ``Master:'' lines to file variables.
This will be done when AUC TeX first try to use the master file.")

(defun TeX-add-local-master ()
  "Add local variable for TeX-master."

  (if (and (buffer-file-name)
           (string-match TeX-one-master
                         (file-name-nondirectory (buffer-file-name)))
           (not buffer-read-only))
      (progn
        (goto-char (point-max))
        (if (re-search-backward (concat "^\\([^\n]+\\)Local " "Variables:")
                                (- (point-max) 3000) t)
            (let ((prefix (TeX-match-buffer 1)))
              (re-search-forward (regexp-quote (concat prefix
                                                        "End:")))
              (beginning-of-line 1)
              (insert prefix "TeX-master: " (prin1-to-string TeX-master) "\n"))
          (forward-line 1)
          (insert "% Local " "Variables: \n"
                  "% mode: " (cond ((eq major-mode 'LaTeX-mode)
				    "latex")
				   ((eq major-mode 'plain-TeX-mode)
				    "tex")
				   (t
				    (substring (symbol-name major-mode) 0 -5)))
		  "\n"
                  "% TeX-master: " (prin1-to-string TeX-master) "\n"
                  "% End: \n")))))

;;; @@ Style Paths

(defvar TeX-format-directory (concat TeX-lisp-directory "format/")
  "*Directory containing information about TeX format packages.
Must end with a slash.")

(defvar TeX-auto-global (concat TeX-lisp-directory "auto/")
  "*Directory containing automatically generated information.
Must end with a slash.

For storing automatic extracted information about the TeX macros
shared by all users of a site.")  

(defvar TeX-style-global (concat TeX-lisp-directory "style/")
  "*Directory containing hand generated TeX information.
Must end with a slash.

These correspond to TeX macros shared by all users of a site.")

(defvar TeX-auto-local "auto/"
  "*Directory containing automatically generated TeX information.
Must end with a slash.

This correspond to TeX macros found in the current directory.")

(defvar TeX-style-local "style/"
  "*Directory containing hand generated TeX information.
Must end with a slash.

These correspond to TeX macros found in the current directory.")

(defvar TeX-macro-private (list (expand-file-name "~/TeX/"))
  "*Directories where you store your personal TeX macros.
Each must end with a slash.")

(defvar TeX-auto-private (concat (car TeX-macro-private) TeX-auto-local)
  "*Directory containing automatically generated information.
Must end with a slash.

These correspond to the personal TeX macros.")

(defvar TeX-style-private (concat (car TeX-macro-private) TeX-style-local)
  "*Directory containing hand generated information.
Must end with a slash.

These correspond to the personal TeX macros.")

(defvar TeX-style-path
  (let ((path))
    (mapcar (function (lambda (file) (if file (setq path (cons file path)))))
	    (list TeX-auto-global
		  TeX-style-global
		  TeX-auto-private
		  TeX-style-private
		  TeX-auto-local
		  TeX-style-local))
    path)
  "*List of directories to search for AUC TeX style files.")

(defvar TeX-check-path '("./")
  "*Directory path to search for dependencies.

If nil, just check the current file.
Used when checking if any files have changed.")

;;; @@ Style Files

(defvar TeX-style-hook-list nil
  "List of TeX style hooks currently loaded.

Each entry is a list where the first element is the name of the style,
and the remaining elements are hooks to be run when that style is
active.")

(defvar TeX-byte-compile t
  "*Not nil means try to byte compile auto files before loading.")

(defun TeX-load-style (style)
  "Search for and load each definition for style in TeX-style-path."
  (if (assoc style TeX-style-hook-list)
      ;; We already found it
      ()
    ;; Insert empty list to mark the fact that we have searched.
    (setq TeX-style-hook-list (cons (list style) TeX-style-hook-list))
    ;; Now check each element of the path
    (mapcar (function
             (lambda (name)
               (let* ((name (if (string-match "/$" name)
                                name
                              (concat name "/")))
                      (el (concat name style ".el"))
                      (elc (concat name style ".elc")))
                 (cond ((and (null TeX-byte-compile)
			     (file-readable-p el))
			(load-file el))
		       ((file-newer-than-file-p el elc)
			(if (not (file-writable-p elc))
			    (load-file el)
			  (byte-compile-file el)
			  (load-file elc)))
		       ((file-readable-p elc)
			(load-file elc))
		       ((file-readable-p el)
			(load-file el))))))
            TeX-style-path)))

(defun TeX-add-style-hook (style hook)
  "Give STYLE yet another HOOK to run."
  (let ((entry (assoc style TeX-style-hook-list)))
    (if (null entry)
        (setq TeX-style-hook-list (cons (list style hook) TeX-style-hook-list))
      (setcdr entry (cons hook (cdr entry))))))

(defun TeX-unload-style (style)
  "Forget that we once loaded STYLE."
  (cond ((null (assoc style TeX-style-hook-list)))
        ((equal (car (car TeX-style-hook-list)) style)
         (setq TeX-style-hook-list (cdr TeX-style-hook-list)))
        (t
         (let ((entry TeX-style-hook-list))
           (while (not (equal (car (car (cdr entry))) style))
             (setq entry (cdr entry)))
           (setcdr entry (cdr (cdr entry)))))))

(defvar TeX-virgin-style (if (and TeX-auto-global
				  (file-directory-p TeX-auto-global))
			     "VIRTEX"
			   "NoVirtexSymbols")
  "Style all documents use, independent on their format package.")

(defvar TeX-format-package nil
  "Format package loaded by VirTeX")

 (make-variable-buffer-local 'TeX-format-package)

(defvar TeX-active-styles nil
  "List of styles currently active in the document.")

 (make-variable-buffer-local 'TeX-active-styles)

(defvar TeX-inhibit-style-hooks nil
  "If not nil, never execute TeX-run-style-hooks.")
 (make-variable-buffer-local 'TeX-inhibit-style-hooks)

(defvar TeX-inhibited-style-hooks nil
  "Style hooks not run because they were inhibited.")
 (make-variable-buffer-local 'TeX-inhibited-style-hooks)

(defun TeX-run-style-hooks (&rest styles)
  "Run the TeX following style hooks."
  (if TeX-inhibit-style-hooks
      (setq TeX-inhibited-style-hooks
            (append TeX-inhibited-style-hooks styles))
    (mapcar (function
             (lambda (style)
               (if (TeX-member style TeX-active-styles 'string-equal) 
                   ()                   ;Avoid recursion.
                 (setq TeX-active-styles
                       (cons style TeX-active-styles))
                 (TeX-load-style style)
                 (mapcar 'funcall
                         (cdr-safe (assoc style TeX-style-hook-list))))))
            styles)))

(defun TeX-run-format-hooks (&rest formats)
  "Run the VirTeX FORMATS packages, but not style hooks."
  (let ((TeX-inhibit-style-hooks t)
	(TeX-style-path (list TeX-format-directory)))
    (mapcar (function
             (lambda (format)
               (if (TeX-member format TeX-active-styles 'string-equal) 
                   ()                   ;Avoid recursion.
                 (setq TeX-active-styles
                       (cons format TeX-active-styles))
                 (TeX-load-style format)
                 (mapcar 'funcall
                         (cdr-safe (assoc format TeX-style-hook-list))))))
            formats)))

(defvar TeX-parse-self t
  "*Parse file after loading it if no style hook is found for it.")
  (make-variable-buffer-local 'TeX-parse-self)

(defvar TeX-style-hook-applied-p nil
  "Nil, unless the style specific hooks have been applied.")
 (make-variable-buffer-local 'TeX-style-hook-applied-p)

(defun TeX-update-style (&optional force)
  "Run style specific hooks for the current document.

Only do this if it has not been done before, or if optional argument
FORCE is not nil."

  (if (or (eq TeX-auto-update 'silent)	; Not a real TeX buffer
	  (and (not force) TeX-style-hook-applied-p))
      ()
    (setq TeX-style-hook-applied-p t)
    (message "Applying style hooks...")
    (apply 'TeX-run-style-hooks (append TeX-inhibited-style-hooks
					(list (TeX-strip-extension nil nil t)
					      (TeX-master-file))))
    (if (and TeX-parse-self
	     (null (cdr-safe (assoc (TeX-strip-extension nil nil t)
				    TeX-style-hook-list))))
	(apply 'TeX-run-style-hooks (TeX-auto-find-files)))
    
    (message "Applying style hooks... done")))

(defvar TeX-remove-style-hooks nil
  "List of hooks to call when we remove the style specific information.")
 (make-variable-buffer-local 'TeX-remove-style-hooks)

(defun TeX-remove-style ()
  "Remnove all style specific information."
  (setq TeX-style-hook-applied-p nil)
  (mapcar 'funcall TeX-remove-style-hooks)
  (setq TeX-active-styles (list TeX-virgin-style TeX-format-package)))

(defun TeX-style-list ()
  "Return a list of all styles (subfils) use by the current document."
  (TeX-update-style)
  TeX-active-styles)

;;; @@ Special Characters

(defvar TeX-esc "\\" "The TeX escape character.")
 (make-variable-buffer-local 'TeX-esc)

(defvar TeX-grop "{" "The TeX group opening character.")
 (make-variable-buffer-local 'TeX-grop)

(defvar TeX-grcl "}" "The TeX group closing character.")
 (make-variable-buffer-local 'TeX-grcl)

;;; @@ Symbols
;;;
;;; Must be before keymaps.

(defvar TeX-complete-word (key-binding "\e\t")
  "*Function to call if M-t is invoked on a word which does not start 
with a backslash.  Default is the meaning of M-t when latex-mode was
first invoked.")

(defvar TeX-symbol-list nil
  "List of known TeX macros.")

 (make-variable-buffer-local 'TeX-symbol-list)

(defvar TeX-symbol-list-changed nil
  "True if the symbol list has changed since last time it was sorted.")

 (make-variable-buffer-local 'TeX-symbol-list-changed)

(defun TeX-add-symbols (&rest symbols)
  "Add SYMBOLS (any number of arguments) to the list of known symbols."
  (if TeX-symbol-list-changed
      (setq TeX-symbol-list (cons symbols TeX-symbol-list))
    (setq TeX-symbol-list-changed t)
    (setq TeX-symbol-list (list symbols TeX-symbol-list))))

(defun TeX-symbol-list ()
  "Return list of symbols TeX knows about.
Each element is a list with the symbol (a string) as the first entry."
  (TeX-update-style)
  (if (null TeX-symbol-list-changed)
      ()
    (setq TeX-symbol-list-changed nil)
    ;; Sort it
    (message "Sorting symbols...")
    (setq TeX-symbol-list
          (sort (mapcar 'TeX-listify (apply 'append TeX-symbol-list))
                'TeX-car-string-lessp))
    ;; Make it unique
    (message "Removing duplicates...") 
    (let ((entry TeX-symbol-list))
      (while (and entry (cdr entry))
        (let ((this (car entry))
              (next (car (cdr entry))))
          (if (not (string-equal (car this) (car next)))
              (setq entry (cdr entry))
            ;; We have to equal symbols.  Use the one with
            ;; most arguments.
            (if (> (length next) (length this))
                (setcdr this (cdr next)))
            (setcdr entry (cdr (cdr entry)))))))
    (message "Removing duplicates... done"))
  TeX-symbol-list)

(setq TeX-remove-style-hooks
      (cons (function (lambda () (setq LaTeX-symbol-list nil)))
            TeX-remove-style-hooks))

;;; @@ The Mode

(defvar TeX-format-list
  '(("JSLITEX" japanese-slitex-mode
     "^[^%\n]*\\\\documentstyle[^%\n]*{jslides}")
    ("SLITEX" slitex-mode "^[^%\n]*\\\\documentstyle[^%\n]*{slides}")
    ("FOILTEX" foiltex-mode "^[^%\n]*\\\\documentstyle[^%\n]*{foils}")
    ("AMSLATEX" ams-latex-mode "^[^%\n]*\\\\documentstyle[^%\n]*{ams")
    ("JLATEX" japanese-latex-mode "^[^%\n]*\\\\documentstyle[^%\n]*{j")
    ("LATEX" latex-mode
     "^[^%\n]*\\\\\\(begin\\|section\\|chapter\\|documentstyle\\)[*{[]")
    ("AMSTEX" ams-tex-mode "-- string likely in AmS TeX --")
    ("JTEX" japanese-plain-tex-mode "-- string likely in Japanese TeX --")
    ("TEX" plain-tex-mode "."))
  "*List of format packages to consider when choosing a TeX mode.

A list with a entry for each format package available at the site.

Each entry is a list with three elements.

1. The name of the format package.
2. The name of the major mode.
3. A regexp typically matched in the beginning of the file.

When entering tex-mode, each regexp is tried in turn in order to find
when major mode to enter.")

(defvar TeX-default-mode 'latex-mode
  "*Mode to enter for a new file when it can't be determined whether
the file is plain TeX or LaTeX or what.")

(defvar TeX-force-default-mode nil
  "*If set to nil, try to infer the mode of the file from its content.")

(defun tex-mode ()
  "Major mode for editing files of input for TeX or LaTeX.
Tries to guess whether this file is for plain TeX or LaTeX.

The algorithm is as follows:

   1) if the file is empty or TeX-force-default-mode is not set to nil, 
      TeX-default-mode is chosen 
   2) If \\documentstyle or \\begin{, \\section{, \\part{ or \\chapter{ is
      found, latex-mode is selected.
   3) Otherwise, use plain-tex-mode "
  (interactive)

  (funcall (if (or (equal (buffer-size) 0)
                   TeX-force-default-mode)
               TeX-default-mode
             (save-excursion
               (goto-char (point-min))
               (let ((entry TeX-format-list)
                     answer)
                 (while (and entry (not answer))
                   (if (re-search-forward (nth 2 (car entry))
                                          nil t)
                       (setq answer (nth 1 (car entry))))
                   (setq entry (cdr entry)))
                 (if answer
                     answer
                   TeX-default-mode))))))

(defun VirTeX-mode (format)
  "Initialize major mode for VirTeX FORMAT package."
  (interactive "s")
  
  (kill-all-local-variables)

  (setq TeX-format-package format)
  (setq local-abbrev-table text-mode-abbrev-table)
  (setq indent-tabs-mode nil)

  ;; Redefine some standard varaibles
  (make-local-variable 'paragraph-start)
  (make-local-variable 'paragraph-separate)
  (make-local-variable 'comment-start)
  (setq comment-start "%")
  (make-local-variable 'comment-start-skip)
  (make-local-variable 'comment-indent-hook)
  (setq comment-indent-hook 'TeX-comment-indent)
  (make-local-variable 'compile-command)
  (if (boundp 'compile-command)
      ()
    (setq compile-command "make"))

  ;; We want this to be early in the list, so we do not add it before
  ;; we enter TeX mode  the first time.
  (or (memq 'TeX-auto-write-hook write-file-hooks)
      (setq write-file-hooks (cons 'TeX-auto-write-hook write-file-hooks)))
  (make-local-variable 'TeX-auto-update)
  (setq TeX-auto-update t)

  ;; Run common hooks
  (TeX-run-format-hooks TeX-virgin-style TeX-format-package))

(defun plain-tex-mode ()
  "Major mode for editing files of input for plain TeX.
Makes $ and } display the characters they match.
Makes \" insert `` when it seems to be the beginning of a quotation,
and '' when it appears to be the end; it inserts \" only after a \\.

See info under AUC TeX for full documentation.

Special commands:
\\{TeX-mode-map}
 
Entering plain-tex-mode calls the value of text-mode-hook,
then the value of TeX-mode-hook, and then the value
of plain-TeX-mode-hook."
  (interactive)
  (VirTeX-mode "TEX"))

(defun ams-tex-mode ()
  "Major mode for editing files of input for AmS TeX.
Makes $ and } display the characters they match.
Makes \" insert `` when it seems to be the beginning of a quotation,
and '' when it appears to be the end; it inserts \" only after a \\.

See info under AUC TeX for full documentation.

Special commands:
\\{TeX-mode-map}
 
Entering ams-tex mode calls the value of text-mode-hook,
then the value of TeX-mode-hook, the value of plain-TeX-mode-hook, and
then the value of AmS-TeX-mode-hook."
  (interactive)
  (VirTeX-mode "AMSTEX"))

(defun latex-mode ()
  "Major mode for editing files of input for LaTeX.

Makes $ and } display the characters they match.
Makes \" insert `` when it seems to be the beginning of a quotation,
and '' when it appears to be the end; it inserts \" only after a \\.
LFD and TAB indent lines as with programming modes.

See under AUC TeX for full documentation.

Special commands:
\\{LaTeX-mode-map}

Entering LaTeX mode calls the value of text-mode-hook,
then the value of TeX-mode-hook, and then the value
of LaTeX-mode-hook."
  (interactive)
  (VirTeX-mode "LATEX"))

(defun slitex-mode ()
  "Major mode for editing files of input for SliTeX.

Makes $ and } display the characters they match.
Makes \" insert `` when it seems to be the beginning of a quotation,
and '' when it appears to be the end; it inserts \" only after a \\.
LFD and TAB indent lines as with programming modes.

See under AUC TeX for full documentation.

Special commands:
\\{LaTeX-mode-map}

Entering LaTeX mode calls the value of text-mode-hook,
then the value of TeX-mode-hook, and then the value
of LaTeX-mode-hook. and then the value of SliTeX-mode-hook."
  (interactive)
  (VirTeX-mode "SLITEX"))

(defun ams-latex-mode ()
  "Major mode for editing files of input for AmS-LaTeX.

Makes $ and } display the characters they match.
Makes \" insert `` when it seems to be the beginning of a quotation,
and '' when it appears to be the end; it inserts \" only after a \\.
LFD and TAB indent lines as with programming modes.

See under AUC TeX for full documentation.

Special commands:
\\{LaTeX-mode-map}

Entering AmS LaTeX mode calls the value of text-mode-hook, then the
value of TeX-mode-hook, then the value of AmS-TeX-mode-hook, and then
the value of LaTeX-mode-hook. and then the value of
AmS-LaTeX-mode-hook." 
  (interactive)
  (VirTeX-mode "AMSLATEX"))

(defun foiltex-mode ()
  "Major mode for editing files of input for FoilTeX.

Makes $ and } display the characters they match.
Makes \" insert `` when it seems to be the beginning of a quotation,
and '' when it appears to be the end; it inserts \" only after a \\.
LFD and TAB indent lines as with programming modes.

FoilTeX is a latex macropackage for creating foils.

FoilTeX has been devoped by IBM, and is available for non-commercial
experimental use.  See the file fltplain.tex of the FoilTeX distribution
for the correct and complete license agreement.

FoilTeX can be copied from many major ftp archives, for example at
ymir.claremont.edu in tex/inputs/foiltex.

Special commands:
\\{LaTeX-mode-map}

Entering FoilTeX mode calls the value of text-mode-hook,
then the value of TeX-mode-hook, and then the value
of LaTeX-mode-hook. and then the value of FoilTeX-mode-hook."
  (interactive)
  (VirTeX-mode "FOILTEX"))

;;; @@ Miscellaneous
;;;
;;; Some of these functions has little to do with TeX, but nonetheless we
;;; should use the "TeX-" prefix to avoid name clashes.

(defvar TeX-auto-regexp-list 'TeX-auto-full-regexp-list
  "*List of regular expresions used for parsing the current file.")
  (make-variable-buffer-local 'TeX-auto-regexp-list)

(defvar TeX-file-extensions '("tex" "sty")
  "*File extensions used by manually generated TeX files.")

(defvar TeX-all-extensions '("[^.\n]+")
  "All possible file extensions.")

(defvar TeX-default-extension "tex"
  "*Default extension for TeX files.")

  (make-variable-buffer-local 'TeX-default-extension)

(defvar BibTeX-file-extensions '("bib")
  "Valid file extensions for BibTeX files.")

(defvar BibTeX-style-extensions '("bst")
  "Valid file extensions for BibTeX styles.")

(defvar TeX-ignore-file "\\(^\\|/\\)\\(\\.\\|\\.\\.\\|RCS\\|SCCS\\|CVS\\)$"
  "*Regular expression matching file names to ignore.

These files or directories will not be considered when searching for
TeX files in a directory.")

(defvar TeX-file-recurse t
  "*If not nil, search TeX directories recursivly.")

(defun TeX-match-extension (file &optional extensions)
  "Return non-nil if FILE has an one of EXTENSIONS.

If EXTENSIONS is not specified or nil, the value of
TeX-file-extensions is used instead."

  (if (null extensions)
      (setq extensions TeX-file-extensions))

  (let ((regexp (concat "\\.\\("
                        (mapconcat 'identity extensions "\\|")
                        "\\)$")))
    (string-match regexp file)))

(defun TeX-strip-extension (&optional string extensions nodir nostrip)
  "Return STRING without any trailing extension in EXTENSIONS.
If NODIR is set, also remove directory part of STRING. 
If NOSTRIP is set, do not remove extension after all.
STRING defaults to the name of the current buffer.
EXTENSIONS defaults to TeX-file-extensions."
  
  (if (null string)
      (setq string (or (buffer-file-name) "<none>")))
  
  (if (null extensions)
      (setq extensions TeX-file-extensions))
  
  (let ((strip (if (and (not nostrip)
                        (TeX-match-extension string extensions))
                   (substring string 0 (match-beginning 0))
                 string)))
    (if nodir
        (file-name-nondirectory strip)
      strip)))

(defun TeX-search-files (&optional directories extensions nodir strip)
  "Return a list of all reachable files in DIRECTORIES ending with EXTENSIONS.
If optional argument NODIR is set, remove directory part.
If optional argument STRIP is set, remove file extension.
If optional argument DIRECTORIES is set, search in those directories. 
Otherwise, search in all TeX macro directories.
If optional argument EXTENSIONS is not set, use TeX-file-extensions"

  (if (null extensions)
      (setq extensions TeX-file-extensions))
  
  (if (null directories)
      (setq directories
	    (cons "./" (append TeX-macro-private TeX-macro-global))))
  
  (let (match)
    
    (while directories
      (let* ((directory (car directories))
             (content (and directory
			   (file-readable-p directory)
			   (file-directory-p directory)
			   (directory-files directory))))
        
        (setq directories (cdr directories))
	
        (while content
          (let ((file (concat directory (car content))))
	    
            (setq content (cdr content))
            (cond ((string-match TeX-ignore-file file))
		  ((not (file-readable-p file)))
                  ((file-directory-p file)
		   (if TeX-file-recurse
		       (setq directories
			     (cons (concat file "/") directories))))
                  ((TeX-match-extension file extensions)
                   (setq match (cons (TeX-strip-extension file
							  extensions
							  nodir
							  (not strip))
                                     match))))))))
    
    match))

(defun TeX-split-string (char string)
  "Returns a list of strings. given REGEXP the STRING is split into 
sections which in string was seperated by REGEXP.

Examples:

      (TeX-split-string \"\:\" \"abc:def:ghi\")
          -> (\"abc\" \"def\" \"ghi\")

      (TeX-split-string \" *\" \"dvips -Plw -p3 -c4 testfile.dvi\")

          -> (\"dvips\" \"-Plw\" \"-p3\" \"-c4\" \"testfile.dvi\")

If CHAR is nil, or \"\", an error will occur."

  (let ((regexp char)
        (start 0)
        (result '()))
    (while (string-match regexp string start)
      (let ((match (string-match regexp string start)))
        (setq result (cons (substring string start match) result))
        (setq start (match-end 0))))
    (setq result (cons (substring string start nil) result))
    (nreverse result)))

(defun TeX-car-string-lessp (a b)
  (string-lessp (car a) (car b)))

(defun TeX-listify (a)
  (if (listp a) a (list a)))

(defun TeX-member (elt list how)
  "Returns the member ELT in LIST.  Comparison done with HOW.

Return nil if ELT is not a member of LIST."
  (while (and list (not (funcall how elt (car list))))
    (setq list (cdr list)))
  (car-safe list))

(defun TeX-match-buffer (n)
  "Return the substring corresponding to the N'th match.

See match-data for details."
  (buffer-substring (match-beginning n) (match-end n)))

(defun TeX-function-p (arg)
  "Return non-nil if ARG is collable as a function."
  (or (and (fboundp 'byte-code-function-p)
	   (byte-code-function-p arg))
      (and (listp arg)
	   (eq (car arg) 'lambda))
      (and (symbolp arg)
	   (fboundp arg))))

;;; @@ Emacs

(run-hooks 'TeX-after-tex-init-hook)

;;; Local Variables:
;;; mode: emacs-lisp
;;; mode: outline-minor
;;; outline-regexp: ";;; @+\\|(......"
;;; End:
