;;; @ tex-math.el - Smart abbrevation for common math expressions.
;;; 
;;; $Id: tex-math.el,v 5.24 1993/07/07 00:41:20 amanda Exp $

(require 'min-bind)

;;; @@ Copyright
;;;
;;; Copyright (C) 1991 Kresten Krab Thorup (krab@iesd.auc.dk).
;;; Copyright (C) 1993 Per Abrahamsen

;;; TO DO LIST
;;;
;;; Make the documentation be updated on changes to the list
;;;
;;; 
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 1, or (at your option)
;;; any later version.
;;; 
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;; 
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;; @@ Customization

(defvar LaTeX-math-abbrev-prefix "`"
  "*Prefix key for use in LaTeX-math-mode.")

;;; @@ The Mode

(defvar TeX-math-mode nil
  "Is LaTeX-math-mode on or off? non nil means on.")

 (make-variable-buffer-local 'TeX-math-mode)

(or (assoc 'TeX-math-mode minor-mode-alist)
    (setq minor-mode-alist (cons '(TeX-math-mode " Math") minor-mode-alist)))

(defun LaTeX-math-mode (&optional arg)
  "A minor mode with easy acces to TeX math macros. 
Further documentation will be present when the mode has been set."
  (interactive "P")
  (if (or (and (null arg) TeX-math-mode)
	  (<= (prefix-numeric-value arg) 0))
      ;; Turn it off
      (if TeX-math-mode
	  (minor-unbind 'TeX-math-mode))
    (if (not TeX-math-mode)
	(let ((map (make-sparse-keymap)))
	  (define-key map LaTeX-math-abbrev-prefix 'LaTeX-abbrev-math)
	  (minor-add-to-keymap 'TeX-math-mode map)
	  (minor-set-variable 'TeX-math-mode 'TeX-math-mode t))))

  (set-docstring
   'LaTeX-math-mode
   (concat
    "A minor mode, in which " LaTeX-math-abbrev-prefix " is"
    " bound to a list of standard math symbols.\n"
    "With prefix (C-u) the symbol is inserted in single dollar signs.\n"
    "The prefix char may be customized by changing the value of the\n"
    "value of the string variable LaTeX-math-abbrev-prefix.\n\n"
    "Current keybindings are:\n\n"
    LaTeX-math-abbrev-prefix " ?   Help (this list)\n"
    (apply (function concat)
	   (mapcar (function
		    (lambda (item)
		      (let ((key (car item))
			    (val (cdr item)))
			(concat LaTeX-math-abbrev-prefix " "
				(char-to-string key) "   "
				(if (symbolp val)
				    (concat "Function: " (symbol-name val))
				  (concat "\\" val))
				"\n"))))
		   LaTeX-mathlist))))

  ;; force modeline update soon
  (set-buffer-modified-p (buffer-modified-p)))

;;; @@ Utility

(minor-if19
    (defun set-docstring (symbol string)
      "Replace the function docstring for SYMBOL with STRING."
      (let* ((symfunc (symbol-function symbol))
	     (doc-cdr (and (listp symfunc) (nthcdr 2 symfunc)))
	     (doc-str (documentation symbol)))
	(if doc-cdr
	    (setcar doc-cdr string)
	  (fset symbol 
		(apply 'make-byte-code
		       (aref symfunc 0) (aref symfunc 1)
		       (aref symfunc 2) (aref symfunc 3)
		       string
		       (if (> (length symfunc) 5)
			   (list (aref symfunc 5))))))))
  (defun set-docstring (symbol string)
    "Set the documentation string for symbol SYMBOL to STRING."
    (let* ((symfunc (symbol-function symbol))
	   (doc-cdr (nthcdr 2 symfunc)))
      (condition-case nil 
	    (setcar doc-cdr string)
	  (error
	   (setq symfunc (copy-sequence (symbol-function
					 symbol))
		 doc-cdr (nthcdr 2 symfunc))
	   (setcar doc-cdr string)
	   (fset symbol symfunc))))))

;;; @@ Data

(defvar LaTeX-mathlist
  '(( ?a . "alpha" )
    ( ?b . "beta" )
    ( ?c . LaTeX-cal)
    ( ?d . "delta" )
    ( ?e . "epsilon" )
    ( ?f . "phi" )
    ( ?g . "gamma" )
    ( ?h . "eta" )
    ( ?k . "kappa" )
    ( ?l . "lambda" )
    ( ?m . "mu" )
    ( ?n . "nu" )
    ( ?o . "omega" )
    ( ?p . "pi" )
    ( ?q . "theta" )
    ( ?r . "rho" )
    ( ?s . "sigma" )
    ( ?t . "tau" )
    ( ?u . "upsilon" )
    ( ?x . "chi" )
    ( ?y . "psi" )
    ( ?z . "zeta" )
    ( ?D . "Delta" )
    ( ?G . "Gamma" )
    ( ?Q . "Theta" )
    ( ?L . "Lambda" )
    ( ?Y . "Psi" )
    ( ?P . "Pi" )
    ( ?S . "Sigma" )
    ( ?U . "Upsilon" )
    ( ?V . "Phi" )
    ( ?O . "Omega" )
    ( ?\C-f . "rightarrow" )
    ( ?\C-b . "leftarrow" )
    ( ?\C-p . "uparrow" )
    ( ?\C-n . "downarrow" )
    ( ?< . "leq" )
    ( ?> . "geq" )
    ( ?~ . "tilde" )
    ( ?I . "infty" )
    ( ?A . "forall" )
    ( ?E . "exists" )
    ( ?! . "not" )
    ( ?i . "in" )
    ( ?* . "times" )
    ( ?. . "cdot" )
    ( ?{ . "subset" )
    ( ?} . "supset" )
    ( ?\[ . "subseteq" )
    ( ?\] . "supseteq" )
    ( ?\\ . "backslash" )
    ( ?/ . "setminus" )
    ( ?+ . "cup" )
    ( ?- . "cap" )
    ( ?\( . "langle")
    ( ?\) . "rangle")
    ( ?\C-e . "exp")
    ( ?\C-s . "sin")
    ( ?\C-c . "cos")  
    ( ?\C-^ . "sup" )
    ( ?\C-_ . "inf" )
    ( ?\C-d . "det" )
    ( ?\C-l . "lim" )
    ( ?\C-t . "tan" )
    ( ?^ . "hat")
    ( ?v . "vee")
    ( ?0 . "emptyset" ))
  "Alist of key-symbol mappings to use in LaTeX-math-mode.
The user may define new mapping by redefining this variable.

To expand the current mapping by binding \"\\alpha\" to `-a you may insert the followin in your .emacs:

   (setq TeX-after-tex-math-hook (function (lambda ()
                                            (setq LaTeX-mathlist
                                                  (cons '( ?a . \"alpha\")
			                           LaTeX-mathlist)))))")

;;; @@ Interactive Commands

(defun LaTeX-abbrev-math (p char)
  "Smart bindings of math symbols. See LaTeX-math-mode for documentation."
  (interactive "*p\nc")
  (let ( (association nil) )
    (cond ((= (string-to-char LaTeX-math-abbrev-prefix) char)
	   (insert-string LaTeX-math-abbrev-prefix))
	  ((= ?? char)
	   (describe-function 'LaTeX-math-mode))
	  ((setq association (assoc char LaTeX-mathlist))
	   (let ( (val (cdr association)) )
	     (if (/= p 1) (insert "$"))
	     (if (stringp val)
		 (insert "\\" val)
	       (call-interactively val))
	     (if (/= p 1) (insert "$"))))
	  (t
	   (error "no such LaTeX-math-abbrev defined")))))

(defun LaTeX-cal (char)
  "Inserts a {\cal CHAR}."
  (interactive "*c")
  (insert "{\\cal " (char-to-string char) "}"))

;;; @@ Emacs

(run-hooks 'TeX-after-tex-math-hook)

;;; Local Variables:
;;; mode: emacs-lisp
;;; mode: outline-minor
;;; outline-regexp: ";;; @+\\|(......"
;;; End:
